-- main file running the bulk of the game and all calculations
--[[

	HALtris, a take on the classic puzzle game Tetris where an AI may either help or
	obstruct you. Copyright Husam Aldahiyat 2013, @numandina.
	
	Function list:

	General:	
		maingame.load()
		setupBoard()
		maingamedraw()
		
	Sprite batch setup:		
		setupTileset()
		updateTilesetBatch()
		
	Piece operations:	
		createBlock()
		canRotate()
		rotate()
		canMove()
		move()
	
	logic checks:	
		isOutside()
		isOccupied()
		isOccupiedSelf()
		
	Dealing with input:	
		love.keypressed()
		isInsideBox()
		lovemousepressed()
	
	Game Flow:
		loveupdate()
		finishTurn()
		checkLines()
		checkGravity()
		goDown()
		createNextBlock()
		
	A.I.:
		gotoAI_1()
		fitness()
		maxHeight()
		meanHeight()
		cavities()
		freeEdges()
		doesLine()
		
	A.I. Simulation:
		copyGrid()
		calcFitness()
		simulateRun()
		
	Other:
		dispStats()
		love.focus()
		love.quit()
]]

-- taking an instance so I can launch it from main.lua
local maingame = {}

-- load function, initializes variables and prepares the canvas for playing. this is
-- executed right after the game launches
function maingame.load()
	
	-- create and reallocate a particle system. we'll use this when a line is cleared
	part1 = love.graphics.newImage("media/part1.png");
	p = love.graphics.newParticleSystem(part1, 1000)
	p:setEmissionRate(1000)
	p:setSpeed(300, 400)
	p:setSizes(1, .5)
	p:setColors(244, 77, 20, 255, 194, 30, 18, 110, 155, 0, 18, 0)
	p:setPosition(200, 500)
	--p:setLifetime(0.25)
	--p:setGravity(1000, 0)
	--p:setParticleLife(0.4)
	--p:setDirection(0)
	--p:setSpread(360)
	--p:setTangentialAcceleration(1000)
	--p:setRadialAcceleration(-2000)
	
	-- do not run this particle system unless told to
	p:stop()

	-- string holding the current state of the game. also displays useful info
	status = ""

	-- load background music to be played later
	backg = love.audio.newSource("media/audio/backg.ogg")
	backg:setLooping(true)
	
	-- load sounds
	shh = love.audio.newSource("media/audio/shh.ogg")
	click = love.audio.newSource("media/audio/click.ogg")
	pling = love.audio.newSource("media/audio/pling.ogg")
	beep = love.audio.newSource("media/audio/beep.ogg")

	-- Constants
	screenW = 800
	screenH = 800
		
	-- block width and height, very important constants
	blockW = 30
	blockH = 30
	
	-- number of rows and columns for playing field. change for a different experience
	-- default is the standard numbers for Tetris
	nRows = 16
	nCols = 10
	
	-- more constants to make things easier later on. these are the x and y origin
	-- for the game field. as seen, the playground is padded 100 pixels from the left 
	-- and another 100 from the bottom
	orgX = 100
	orgY = screenH - (blockH * nRows + 100)	

	-- Image Definitions
	HAL = love.graphics.newImage("media/HAL9000.png")
	
	-- stored in an array, for better referencing
	blocks = {block_1, block_2, block_3, block_4, block_5, block_6, block_7}

	-- playing field, matrix. initially empty. we create a mathematical grid, a matrix,
	-- which will hold the data for our playing field. this matrix is where all work will
	-- be held, and the graphical display is nothing more than a graphical representation
	-- of this grid. 
	-- we create 4 additional rows to account for pieces right at the start of each turn
	grid = {}
	for i = 1, nRows + 4 do
		grid[i] = {}
		for j = 1, nCols do
			grid[i][j] = 0
		end
	end
	
	-- create a definition for the piece currently controlled by the player. of its
	-- properties are whether or not it's active (as opposed to settled), its rotation,
	-- type, and of course the coordinates of each of the four block composing it
	curBlock = {
		active = false,
		rotation = 1;
		btype = 0,
		co1 = {x = 0, y = 0},
		co2 = {x = 0, y = 0},
		co3 = {x = 0, y = 0},
		co4 = {x = 0, y = 0},
	}

	-- variables for the right-hand scoring display
	sc_SPEED = 1
	sc_LINES = 0
	sc_SCORE = 0
	sc_COMBO = 0

	-- variables for the metrics used when categorizing how "orderly" a grid is or how
	-- much potential a piece has. we'll get to these in more detail up ahead
	d1 = 0
	d2 = 0
	d3 = 0
	d4 = 0
	d5 = 0
	d6 = "No"
	d7 = 0
	d8 = 0
	
	-- timer responsible for pulling down the controlled piece during gameplay
	gravitytimer = 0
	
	-- loading a file to store the high score in. if no file found, create a new one.
	fname = "high_sc"
	if not love.filesystem.exists(fname) then
		love.filesystem.write(fname, 0)
	end
	
	-- VERY IMPORTANT! the weights/coefficients for the metrics, responsible for setting
	-- how valuable a piece is, based on how complex the playing field is
	c1 = 5
	c2 = 10
	c3 = 3
	c4 = 3
	c8 = 3
	c5 = .5
	c6 = -1000
	
	-- create the tile set responsible for block graphics
	setupTileset()
	
	-- create created as the AI testbed and hypothetical scenario simulator
	newGrid = {}
	for i = 1, nRows + 4 do
		newGrid[i] = {}
		for j = 1, nCols do
			newGrid[i][j] = 0
		end
	end	
	
	-- counter to be used in a later loop (run simulation)
	jG = 0
	
	-- bolean corresponding to the on/off simulation button
	canSimulate = true
	
	-- used to store fitness values for each piece	
	V = {}

	-- initialize for display	
	for i = 1, 7 do
		V[i] = 0
	end


end

-- board setup. this is executed when specifically told you, when we're ready to start
-- playing
function setupBoard()

	-- start playing background music
	love.audio.play(backg)
	
	-- create the first piece
	createNextBlock()		
	
	-- change screen resolution to better fit playing field
	love.window.setMode(1500, 800, {resizable=true, vsync=true, minwidth=500, minheight=266})
end

-- sprite sheet setup
function setupTileset()
	
	-- this will contain all of our quads. each quad is a small part of the whole
	-- sprite sheet, and will represent one element (block) used for graphical display
	tileQuads = {}
	
	-- take our sprite sheet, which contains all of our block graphics in one image
	tilesetImage = love.graphics.newImage("media/pieces/blocks_sb.png")

	-- empty block
	tileQuads[0] = love.graphics.newQuad(blockW * 1, blockH * 1, blockW, blockH,
	tilesetImage:getWidth(), tilesetImage:getHeight())
	
	-- I block
	tileQuads[1] = love.graphics.newQuad(blockW * 3, blockH * 0, blockW, blockH,
	tilesetImage:getWidth(), tilesetImage:getHeight())
	
	-- J block
	tileQuads[2] = love.graphics.newQuad(blockW * 2, blockH * 0, blockW, blockH,
	tilesetImage:getWidth(), tilesetImage:getHeight())
	
	-- S block
	tileQuads[3] = love.graphics.newQuad(blockW * 0, blockH * 2, blockW, blockH,
	tilesetImage:getWidth(), tilesetImage:getHeight())
	
	-- Z block
	tileQuads[4] = love.graphics.newQuad(blockW * 0, blockH * 0, blockW, blockH,
	tilesetImage:getWidth(), tilesetImage:getHeight())
	
	-- L block
	tileQuads[5] = love.graphics.newQuad(blockW * 1, blockH * 0, blockW, blockH,
	tilesetImage:getWidth(), tilesetImage:getHeight())
	
	-- O block
	tileQuads[6] = love.graphics.newQuad(blockW * 0, blockH * 3, blockW, blockH,
	tilesetImage:getWidth(), tilesetImage:getHeight())
	
	-- T block
	tileQuads[7] = love.graphics.newQuad(blockW * 0, blockH * 1, blockW, blockH,
	tilesetImage:getWidth(), tilesetImage:getHeight())
	
	-- T block
	tileQuads[8] = love.graphics.newQuad(blockW * 1, blockH * 2, blockW, blockH,
	tilesetImage:getWidth(), tilesetImage:getHeight())
	
	-- border
	tileQuads[9] = love.graphics.newQuad(blockW * 1, blockH * 2, blockW, blockH,
	tilesetImage:getWidth(), tilesetImage:getHeight())
	
	-- create our sprite batch!
	tilesetBatch = love.graphics.newSpriteBatch(tilesetImage, 500 * 2)
	
end

-- update our sprite batch continuously, translating the elements of our main matrix GRID
-- into coordinates to be used for graphical display in the draw() function
function updateTilesetBatch()
	
	-- clear in order to be filled with new values
	tilesetBatch:clear()
	
	-- loop through all columns
	for x = 1, nCols do
		
		-- loop through all rows
		for y = 1, nRows + 4 do
			
			-- top four rows should only show the new piece and not any empty blocks 
			if y > 4 or grid[y][x] > 0 then
			
				-- based on the grid's element (a number from 0 - 8) a block image will
				-- be used, where 0 is an empty block and 1 - 7 are different block colors
				tilesetBatch:add(tileQuads[grid[y][x]], 
					orgX + blockW * (x - 1), orgY + blockH * (y - 5))
					
			end
		end
	end
	
	-- do the same for AI playground
	for x = 1, nCols do
		
		-- loop through all rows
		for y = 1, nRows + 4 do
			
			-- top four rows should only show the new piece and not any empty blocks 
			if y > 4 or newGrid[y][x] > 0 then
			
				-- based on the grid's element (a number from 0 - 8) a block image will
				-- be used, where 0 is an empty block and 1 - 7 are different block colors
				tilesetBatch:add(tileQuads[newGrid[y][x]], 
					800 + orgX + blockW * (x - 1), orgY + blockH * (y - 5))
					
			end
		end
	end
	
	-- plot left/right borders. loop through the number of rows while plotting blocks
	for i = 0, nRows do
	
		tilesetBatch:add(tileQuads[8], 
			orgX - blockW, orgY + i * blockH)
			
		tilesetBatch:add(tileQuads[8], 
			orgX + nCols * blockW, orgY + i * blockH)
	end	
	
	-- creating the lower border
	for i = 0, nCols + 1 do
	
		tilesetBatch:add(tileQuads[8], 
			orgX + blockW * (i - 1), orgY + nRows * blockH)
	end
	
	-- create grid for the pieces displayed on the right
	HUDgrid = {
		{0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
		{0, 0, 1, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 7, 0},
		{0, 0, 1, 0, 0, 2, 0, 0, 3, 3, 0, 4, 4, 0, 0, 5, 0, 0, 6, 6, 0, 7, 7},
		{0, 0, 1, 0, 2, 2, 0, 3, 3, 0, 0, 0, 4, 4, 0, 5, 5, 0, 6, 6, 0, 7, 0},
	}
	
	-- reference location for these static pieces
	orgXHUD = 800
	orgYHUD = 0
	
	-- plot this grid
	for x = 1, 23 do
		for y = 1, 4 do
			if HUDgrid[y][x] > 0 then
				tilesetBatch:add(tileQuads[HUDgrid[y][x]], 
					orgXHUD + blockW * (x - 1), orgYHUD + blockH * (y - 1))
			end
		end
	end
end

-- draw function, very important
function maingamedraw()

	-- set up background color and font size
	love.graphics.setBackgroundColor(245/1.5, 241/1.5, 222/1.5, 255)
    local f = love.graphics.newFont(12)
    love.graphics.setFont(f)

	-- some graphics on the upper right hand side
 	love.graphics.rectangle("line", 500, 100, 100, 50)
 	love.graphics.rectangle("line", 650, 100, 100, 50)    
 	love.graphics.rectangle("line", 980, 160, 100, 50)    
    love.graphics.print("Pause", 530, 117)
    love.graphics.print("Menu", 683, 117)
    love.graphics.print("You can change difficulty using A and D", 500, 200)
    if canSimulate then
	    love.graphics.print("Turn Off", 1000, 180)
	else
	    love.graphics.print("Turn On", 1000, 180)
	end
	
	-- fitness criterea on the upper left side
    love.graphics.print("Max Height: " ..d1 .." (" ..c1 ..")", 100, 25)
    love.graphics.print("Mean Height: " ..string.format('%1.3f', d2) .." (" ..c2 ..")", 100, 50)
    love.graphics.print("Semi Cavities: " ..d3 .." (" ..c3 ..")", 100, 75)
    love.graphics.print("Cavities: " ..d4 .." (" ..c4 ..")", 100, 100)
    love.graphics.print("Closed: " ..d8 .." (" ..c8 ..")", 100, 125)
    love.graphics.print("Free Edges: " ..d5 .." (" ..c5 ..")", 100, 150)
    love.graphics.print("Can Complete? " ..d6 .." (" ..c6 ..")", 100, 175)
    love.graphics.print("Fitness: " ..string.format('%1.3f', d7), 100, 200)
    
    -- best fitness values for each piece    
    love.graphics.print("I: " ..string.format('%1.3f', V[1]), 850, 125)
    love.graphics.print("J: " ..string.format('%1.3f', V[2]), 935, 125)
    love.graphics.print("S: " ..string.format('%1.3f', V[3]), 1025, 125)
    love.graphics.print("Z: " ..string.format('%1.3f', V[4]), 1130, 125)
    love.graphics.print("L: " ..string.format('%1.3f', V[5]), 1250, 125)
    love.graphics.print("O: " ..string.format('%1.3f', V[6]), 1335, 125)
    love.graphics.print("T: " ..string.format('%1.3f', V[7]), 1430, 125)

	-- Upper Box
 	love.graphics.rectangle("line", orgX + blockW * nCols + 100, orgY, 250, 250)    
    love.graphics.print("Next Piece", orgX + blockW * nCols + 100 + 5, orgY + 5)
	love.graphics.draw(HAL, orgX + blockW * nCols + 100 + (250 - 220)/2, orgY + (250 - 220)/2)
    
    -- Lower Box Constants
    lowX = orgX + blockW * nCols + 100
    lowY = orgY + 250 + 25
    pad1 = 20
	pad2 = 40
	
    -- Lower Box
 	love.graphics.rectangle("line", lowX, lowY, 250, nRows * blockH - 250)    
    love.graphics.print("Difficulty: ", lowX + 5, lowY + 5)    
	love.graphics.print(sc_DIFF, lowX + 5 + pad1, lowY + 5 + pad1)
    love.graphics.print("Speed: ", lowX + 5 + pad2 * 3, lowY + 5)    
	love.graphics.print(sc_SPEED, lowX + 5 + pad1 + pad2 * 3, lowY + 5 + pad1)
    love.graphics.print("Number of Completed Lines: ", lowX + 5, lowY + 5 + pad2 + pad1)
	love.graphics.print(sc_LINES, lowX + 5 + pad1, lowY + 5 + pad1 * 2 + pad2)
    love.graphics.print("Longest Combo: ", lowX + 5, lowY + 5 + pad2 * 2 + pad1 * 2)
	love.graphics.print(sc_COMBO, lowX + 5 + pad1, lowY + 5 + pad1 * 3 + pad2 * 2)
    love.graphics.print("Score: ", lowX + 5, lowY + 5 + pad2 * 3 + pad1 * 3)
	love.graphics.print(sc_SCORE, lowX + 5 + pad1, lowY + 5 + pad1 * 4 + pad2 * 3)
    love.graphics.print("High: ", lowX + 5 + pad2 * 3, lowY + 5 + pad2 * 3 + pad1 * 3)
	love.graphics.print(love.filesystem.read(fname), lowX + 5 + pad1 + pad2 * 3, lowY + 5 + pad1 * 4 + pad2 * 3)
    
	-- display the current game status
	love.graphics.print(status, 375, 200)
	
	-- update the tile set based on current game grid matrix
    updateTilesetBatch()
    
    -- after updating the tile batch, draw it in one go!
	love.graphics.draw(tilesetBatch)

	-- plotting the particle effect
	love.graphics.draw(p, 0, 0)
	
	-- display game FPS
	love.graphics.print("FPS: "..love.timer.getFPS(), 25, screenH - 25)

end

-- creates a block. input can be either a number or a string specifying which piece we want
function createBlock(t, n)

	curBlock.rotation = 1
		
	if t == "I" or n == 1 then
		curBlock.btype = 1
		status = "I-block"
		curBlock.co = {
			{x = 5, y = 1},
			{x = 5, y = 2},
			{x = 5, y = 3},
			{x = 5, y = 4},
		}		
	elseif t == "J" or n == 2 then
		curBlock.btype = 2
		status = "J-block"
		curBlock.co = {
			{x = 6, y = 1},
			{x = 6, y = 2},
			{x = 6, y = 3},
			{x = 5, y = 3},
		}
	elseif t == "S" or n == 3 then
		curBlock.btype = 3
		status = "S-block"
		curBlock.co = {
			{x = 6, y = 2},
			{x = 5, y = 2},
			{x = 5, y = 3},
			{x = 4, y = 3},
		}
	elseif t == "Z" or n == 4 then
		curBlock.btype = 4
		status = "Z-block"
		curBlock.co = {
			{x = 5, y = 2},
			{x = 6, y = 2},
			{x = 6, y = 3},
			{x = 7, y = 3},
		}
	elseif t == "L" or n == 5 then
		curBlock.btype = 5
		status = "L-block"
		curBlock.co = {
			{x = 5, y = 1},
			{x = 5, y = 2},
			{x = 5, y = 3},
			{x = 6, y = 3},
		}
	elseif t == "O" or n == 6 then
		curBlock.btype = 6
		status = "O-block"
		curBlock.co = {
			{x = 5, y = 2},
			{x = 5, y = 3},
			{x = 6, y = 2},
			{x = 6, y = 3},
		}
	elseif t == "T" or n == 7 then
		curBlock.btype = 7
		status = "T-block"
		curBlock.co = {
			{x = 4, y = 2},
			{x = 5, y = 2},
			{x = 6, y = 2},
			{x = 5, y = 3},
		}
	end
	
	if simMode then status = "Calculating" end

end

-- rotates a piece on the grid. the grid is input because we have two grids, one for
-- playing in, and another to be used as a testing template for the AI.
-- whenever we use rotate or move, we mean the piece controlled by the player, i.e.
-- curBlock, mentioned before
function rotate (grid)

	-- we move it by 0, 0 simply to clear the previous (current) position, soon to be
	-- replaced by the new, rotated one
	move(0, 0, grid)
	
	-- change the coordinates of the four blocks constituting the piece. these new
	-- coordinates are already prepared from the function canRotate(), which should
	-- always be executed before rotate()
	curBlock.co = newCo

end

-- check if curBlock is allowed to rotate
function canRotate(grid)
	
	-- take the original coordinates before rotation
	newCo = curBlock.co
	
	-- take the first block of the piece's 4. use it as reference
	ref = newCo[1]
	
---------------- I block --------------------------------
	if curBlock.btype == 1 then
	
		-- Horizontal
		if curBlock.rotation == 1 then
			newCo = {
				{x = ref.x - 1, y = ref.y + 1},
				{x = ref.x - 0, y = ref.y + 1},
				{x = ref.x + 1, y = ref.y + 1},
				{x = ref.x + 2, y = ref.y + 1},
			}			
			newRot = 2
			
		-- Vertical
		elseif curBlock.rotation == 2 then
			newCo = {
				{x = ref.x + 1, y = ref.y - 1},
				{x = ref.x + 1, y = ref.y + 0},
				{x = ref.x + 1, y = ref.y + 1},
				{x = ref.x + 1, y = ref.y + 2},
			}			
			newRot = 1			
		end
		
---------------- J block --------------------------------
	elseif curBlock.btype == 2 then
	
		-- J shape
		if curBlock.rotation == 1 then
			newCo = {
				{x = ref.x + 1, y = ref.y + 1},
				{x = ref.x - 0, y = ref.y + 1},
				{x = ref.x - 1, y = ref.y + 1},
				{x = ref.x - 1, y = ref.y + 0},
			}			
			newRot = 2
			
		-- c__ shape
		elseif curBlock.rotation == 2 then
			newCo = {
				{x = ref.x - 1, y = ref.y + 1},
				{x = ref.x - 1, y = ref.y + 0},
				{x = ref.x - 1, y = ref.y - 1},
				{x = ref.x + 0, y = ref.y - 1},
			}			
			newRot = 3
			
		-- p shape
		elseif curBlock.rotation == 3 then
			newCo = {
				{x = ref.x - 1, y = ref.y - 1},
				{x = ref.x + 0, y = ref.y - 1},
				{x = ref.x + 1, y = ref.y - 1},
				{x = ref.x + 1, y = ref.y + 0},
			}			
			newRot = 4
			
		-- --. shape
		elseif curBlock.rotation == 4 then
			newCo = {
				{x = ref.x + 1, y = ref.y - 1},
				{x = ref.x + 1, y = ref.y + 0},
				{x = ref.x + 1, y = ref.y + 1},
				{x = ref.x + 0, y = ref.y + 1},
			}			
			newRot = 1
		end

---------------- S block --------------------------------
	elseif curBlock.btype == 3 then
	
		-- S shape
		if curBlock.rotation == 1 then
			newCo = {
				{x = ref.x - 1, y = ref.y - 1},
				{x = ref.x - 1, y = ref.y + 0},
				{x = ref.x + 0, y = ref.y + 0},
				{x = ref.x + 0, y = ref.y + 1},
			}			
			newRot = 2
			
		-- 4 shape
		elseif curBlock.rotation == 2 then
			newCo = {
				{x = ref.x + 1, y = ref.y + 1},
				{x = ref.x + 0, y = ref.y + 1},
				{x = ref.x + 0, y = ref.y + 2},
				{x = ref.x - 1, y = ref.y + 2},
			}			
			newRot = 1
		end

---------------- Z block --------------------------------
	elseif curBlock.btype == 4 then
	
		-- Z shape
		if curBlock.rotation == 1 then
			newCo = {
				{x = ref.x + 1, y = ref.y - 1},
				{x = ref.x + 1, y = ref.y + 0},
				{x = ref.x + 0, y = ref.y + 0},
				{x = ref.x + 0, y = ref.y + 1},
			}			
			newRot = 2
			
		-- u shape
		elseif curBlock.rotation == 2 then
			newCo = {
				{x = ref.x - 1, y = ref.y + 1},
				{x = ref.x + 0, y = ref.y + 1},
				{x = ref.x + 0, y = ref.y + 2},
				{x = ref.x + 1, y = ref.y + 2},
			}			
			newRot = 1
		end
		
---------------- L block --------------------------------
	elseif curBlock.btype == 5 then
	
		-- L shape
		if curBlock.rotation == 1 then
			newCo = {
				{x = ref.x + 1, y = ref.y + 1},
				{x = ref.x + 0, y = ref.y + 1},
				{x = ref.x - 1, y = ref.y + 1},
				{x = ref.x - 1, y = ref.y + 2},
			}			
			newRot = 2
			
		-- o-- shape
		elseif curBlock.rotation == 2 then
			newCo = {
				{x = ref.x - 1, y = ref.y + 1},
				{x = ref.x - 1, y = ref.y + 0},
				{x = ref.x - 1, y = ref.y - 1},
				{x = ref.x - 2, y = ref.y - 1},
			}			
			newRot = 3
			
		-- setta shape
		elseif curBlock.rotation == 3 then
			newCo = {
				{x = ref.x - 1, y = ref.y - 1},
				{x = ref.x + 0, y = ref.y - 1},
				{x = ref.x + 1, y = ref.y - 1},
				{x = ref.x + 1, y = ref.y - 2},
			}			
			newRot = 4
			
		-- __o shape
		elseif curBlock.rotation == 4 then
			newCo = {
				{x = ref.x + 1, y = ref.y - 1},
				{x = ref.x + 1, y = ref.y + 0},
				{x = ref.x + 1, y = ref.y + 1},
				{x = ref.x + 2, y = ref.y + 1},
			}			
			newRot = 1
		end
		
---------------- T block --------------------------------
	elseif curBlock.btype == 7 then
	
		-- T shape
		if curBlock.rotation == 1 then
			newCo = {
				{x = ref.x + 1, y = ref.y - 1},
				{x = ref.x + 1, y = ref.y + 0},
				{x = ref.x + 1, y = ref.y + 1},
				{x = ref.x + 0, y = ref.y + 0},
			}			
			newRot = 2
			
		-- -| shape
		elseif curBlock.rotation == 2 then
			newCo = {
				{x = ref.x + 1, y = ref.y + 1},
				{x = ref.x + 0, y = ref.y + 1},
				{x = ref.x - 1, y = ref.y + 1},
				{x = ref.x + 0, y = ref.y + 0},
			}			
			newRot = 3
			
		-- _|_ shape
		elseif curBlock.rotation == 3 then
			newCo = {
				{x = ref.x - 1, y = ref.y + 1},
				{x = ref.x - 1, y = ref.y + 0},
				{x = ref.x - 1, y = ref.y - 1},
				{x = ref.x + 0, y = ref.y + 0},
			}			
			newRot = 4
			
		-- |- shape
		elseif curBlock.rotation == 4 then
			newCo = {
				{x = ref.x - 1, y = ref.y - 1},
				{x = ref.x + 0, y = ref.y - 1},
				{x = ref.x + 1, y = ref.y - 1},
				{x = ref.x + 0, y = ref.y + 0},
			}			
			newRot = 1
		end
		
	end
	
	-- after preparing the new coordinates for each of the four block of the piece,
	-- whichever it was, we test if any of these new position collide with an already
	-- present block
	
	-- loop through the four blocks
	for i = 1, 4 do
		
		-- get the x and y positions
		newX = newCo[i].x
		newY = newCo[i].y

		-- check if new block position lies outside the board
		if isOutside(newX, newY) then
			return false		
		end
		
		-- check if new block position is already occupied by a different piece
		if isOccupied(newY, newX, grid) and not isOccupiedSelf(newX, newY, newCo) then
			return false
		end
	end
	
	-- if we survive to make it here, then we are allowed to rotate. save the new piece
	-- coordinates and return true
	curBlock.rotation = newRot
	
	return true

end

-- function that returns whether or not a coordinate lies out of our playing field
function isOutside(x, y)
	return x > nCols or x < 1 or y > nRows + 4 or y < 1
end

-- function that returns whether a coordinate is occupied in the grid/matrix
function isOccupied(x, y, pGrid)
	return pGrid[x][y] > 0
end

-- takes x and y position as well as piece coordinates. returns whether this position is
-- occupied by one of the piece's blocks itself
function isOccupiedSelf(x, y, Co)
	for j = 1, 4 do
		if (x == Co[j].x) and (y == Co[j].y) then
			return true
		end
	end
	return false
end

-- checks if a piece (curBlock) is allowed to move to a certain position on a grid
function canMove(x, y, grid)
	
	-- loop through all four of the piece's blocks
	for i = 1, 4 do
		
		-- get the new x and y locations for the block's proposed new location
		newX = curBlock.co[i].x + x
		newY = curBlock.co[i].y + y
	
		-- check if this new location lies outside our board
		if isOutside(newX, newY) then
			return false		
		end

		-- check if this new location has already been occupied by another piece
		if isOccupied(newY, newX, grid) and not isOccupiedSelf(newX, newY, curBlock.co) then
			return false
		end
						
	end

	-- if we make it to here then we are allowed to move
	return true
end

-- moves curBlock to a location x,y on grid pGrid. make sure you check canMove() before
function move(x, y, pGrid)
	
	-- loop through the piece's four blocks
	for i = 1, 4 do 
		
		-- clear the block's previous location from our grid
		pGrid[curBlock.co[i].y][curBlock.co[i].x] = 0
		
		-- update the coordinates of the current piece
		-- Note: updating the grid to take into account the piece's location
		-- happens in the update function
		curBlock.co[i].x = curBlock.co[i].x + x
		curBlock.co[i].y = curBlock.co[i].y + y

	end
	
end

-- get keyboard input
function love.keypressed(key, unicode)
	
	-- increase/decrease difficulty through A and D buttons
    if key == 'a' and sc_DIFF > -1 then
		sc_DIFF = sc_DIFF - 1
	end
	
    if key == 'd' and sc_DIFF < 1 then
		sc_DIFF = sc_DIFF + 1
	end

	-- if there is no currently active piece, don't continue to arrow controls. same
	-- if game is paused
	if (not curBlock.active) or gameIsPaused then
		return
	end
   
   -- get arrow keys and check if the move is legal and if so, move or rotate piece. also
   -- make sounds and in the case of pressing down, override the gravity timer
   if key == 'right' then   		
		love.audio.play(click)
		if canMove(1, 0, grid) then
			move(1, 0, grid)
		end
   elseif key == 'left' then	
		love.audio.play(click)
		if canMove(-1, 0, grid) then
			move(-1, 0, grid)
		end
   elseif key == 'down' then	
		love.audio.play(click)
		if canMove(0, 1, grid) then
			gravitytimer = 0
			move(0, 1, grid)
		else
			gravitytimer = 1
		end
   elseif key == 'up' then
		love.audio.play(click)
		if canRotate(grid) then
			rotate(grid)
		end
   end
end

-- the update function, runs every frame, courtesy of love.update from main.lua
function loveupdate(dt)
			  
   -- returns and changes status if game is paused/unpaused
    if gameIsPaused then     
		status = "Paused"
    	return 
	elseif status == "Paused" then
 	    status = "Resumed"
    end

	-- create a force that periodically pulls the current piece down. the time delay
	-- depends on the variable sc_SPEED, which is the game's speed
	
	-- increment the gravity timer
	gravitytimer = gravitytimer + dt
	
	-- increment the simulation timer
	simTimer = simTimer + dt
	
	-- showing the simulation on the AI grid
	if simMode and simTimer > simSpeed then
		simulateRun()
	end
	
	-- if it's time to pull
	if gravitytimer > 1/sc_SPEED then

		-- reset the timer for next time
		gravitytimer = 0
		
		-- check if a movement to below is allowed
		if curBlock.active and canMove(0, 1, grid) then
			move(0, 1, grid)		
		-- if not that means we have reached the bottom. prepare for a new turn
		elseif curBlock.active then
			curBlock.active = false
			finishTurn()
		end
		
	end
	
	-- boolean indicating whether a piece has "crossed" from its initial y position into
	-- the playing field. used for one purpose: to allow rapid fall when holding the
	-- down button. it is set to true unless proven otherwise
	isBeginning = true
	
	-- next is a very important part: updating the grid based on the current moving piece's
	-- block coordinates
	
	-- loop through all of our piece's four blocks
	for i = 1, 4 do
	
		-- if a real piece and not an AI imagined one
		if curBlock.active then
			
			-- update grid, make the matrix's element match the block's type
			grid[curBlock.co[i].y][curBlock.co[i].x] = curBlock.btype		
			
			-- check if any of te four piece's have crossed into the playing field
			if curBlock.co[i].y > 4 then
				isBeginning = false
			end
			
		else
			-- update second grid, make the matrix's element match the block's type
			newGrid[curBlock.co[i].y][curBlock.co[i].x] = curBlock.btype		
		end
	end	
	
	-- if down button is held, and it's not at the begging, increase force of down pull.
	if love.keyboard.isDown("down") and not isBeginning then
		gravitytimer = gravitytimer + .5
	end
	
	-- part 2 of the simulation.
	if simMode and simTimer > simSpeed  then
		calcFitness()
		simTimer = 0
	-- if not simulating create mirror
	elseif not simMode then
		for y = 5, #grid do
			newGrid[y] = {}
			for x = 1, #grid[y] do
				newGrid[y][x] = grid[y][x]
			end
		end		
	end
	
	-- update the particle effect
	p:update(dt)
	
end

-- function that checks if a position lies within a box. used for interacting with
-- buttons with mouse clicks
function isInsideBox(mx, my, x, y, w, h)
    return mx < x + w and mx > x and my > y and y < y + h
end

-- get mouse input, checks if one of the two buttons onscreen have been clicked on
function lovemousepressed(x, y, button)

	if button == 'l' then

		-- if clicked pause button
		if isInsideBox(x, y, 500, 100, 100, 50) then
			gameIsPaused = not gameIsPaused

		-- if clicked Main Menu button. clear graphics, stop music, and change back
		-- window resolution. also revert the isPlaying boolean to its original state
		elseif isInsideBox(x, y, 650, 100, 100, 50) then
			love.graphics.clear()
			love.audio.stop(backg)     	
			love.window.setMode(800, 600, {resizable=true, vsync=true, minwidth=266, minheight=200})
			isPlaying = false
			
		-- Turning on/off the simulation grid
		elseif isInsideBox(x, y, 980, 160, 100, 50) then
			canSimulate = not canSimulate
		end
	end
end

-- function that runs when a block can no longer move downwards, signifying the end of a
-- turn
function finishTurn()
	
	-- increase score by a small amount
	sc_SCORE = sc_SCORE + 10
	
	-- initializing some values so we know if this turn resulted in anything fruitful
	newLines = false
	wentDown = false
	combo = 0
	
	-- check to see if there are any full lines on the board
	checkLines()
	
	-- check to see if the new combo variable is higher than our max combo stat. if
	-- this is the case update with new value
	if sc_COMBO < combo then
		sc_COMBO = combo
	end
	
	-- check if current score is higher than high score. if so update and save to file
	if (0 + love.filesystem.read(fname) < sc_SCORE) then
		love.filesystem.write(fname, sc_SCORE)
	end
	
	-- end game condition. check if there is but a single block on one of the blocks
	-- outside (above) our playing field
	local sum = 0
	for j = 1, nCols do
		sum = sum + grid[4][j]
	end
	
	-- if game has not ended
	if sum == 0 then
		
		-- update game speed based on score
		if sc_SPEED * sc_SPEED * 1000 < sc_SCORE then
			sc_SPEED = sc_SPEED + 1
		end
		
		-- create next block
		createNextBlock()
		
	-- if game has ended
	else
		status  = "Game Over"
	end
	
	-- display the stats on the upper left corner
	dispStats()
	
end

-- check if there are any full lines on the playing board
function checkLines()

	-- loop through all rows
	for i = 5, nRows + 4 do
	
		-- check if in any one row, all the columns satisfy x > 0
		local sum = true
		for j = 1, nCols do
			sum = sum and grid[i][j] > 0
		end
		
		-- if so
		if sum then
			
			-- clear this line
			for j = 1, nCols do
				grid[i][j] = 0
			end
			
			-- indicate that there has indeed been a fruitful change after this turn
			newLines = true
			
			-- increase combo counter
			combo = combo + 1
			
			-- increase number of lines counter
			sc_LINES = sc_LINES + 1
			
			-- increase score
			sc_SCORE = sc_SCORE + 100 * combo * (sc_DIFF + 1) * (sc_DIFF + 1) * (sc_DIFF + 1)
			
			-- show particle effect at the location of the line
			p:setPosition(orgX + blockW * 5, orgY + (i - 4) * blockH)
			p:start()
			
			-- check if there are any other lines
			checkLines()
			return
		end
	end
	
	-- if there had been a change (fruitful change, remember?) apply gravity check function
	if newLines then
		newLines = false
		checkGravity()
	end
	
end

-- checks if there are any blocks above a cleared line. if so, lower all blocks above said
-- line
function checkGravity()
	
	-- we create a boolean to count when we encounter our first block. we can only search
	-- for cleared lines beneath an already placed block, reason being the entire board
	-- may be filled with empty lines (case in point at the start) without there being any
	-- need to apply gravity
	beg = false
	
	-- loop through rows
	for i = 6, nRows + 4 do
	
		-- check if any row has but one block in it
		local sum = 0
		for j = 1, nCols do
			sum = sum + grid[i][j]
		end
		
		-- if yes, start checking for empty lines from there
		if sum > 0 then
			beg = true
		end
		
		-- if there is indeed an empty line after the first block, execute the gravity
		-- function goDown() from a certain row and up
		if sum == 0 and beg then
			wentDown = true
			goDown(i)
			return
		end
	end
	
	-- again, if there has been any change (fruitful), recheck if anything new has transpired
	-- due to our landscape changing actions
	if wentDown then
		wentDown = false
		checkLines()
	end
	
end

-- function that lowers all rows above a certain row number. used for gravity in case of
-- line clearances below other blocks
function goDown(x)

	for i = x, 5, -1 do
		for j = 1, nCols do
			grid[i][j] = grid[i - 1][j]
		end
	end	
	
	-- play cool explosion sound
	love.audio.play(shh)
	
	-- check if there is a need to recall this function again
	checkGravity()

end

-- function that handles creating a new block after a turn has ended, or before a game has
-- started 
function createNextBlock()
	
	-- change status, just in case the computer is slow enough to display it
	status = "Calculating"

	-- if difficulty is 0 (neutral AI), create a block at random
	if sc_DIFF == 0 then
		i = math.random(1, 7)
		createBlock("", i)
		curBlock.active = true
		
	-- difficulty = -1, execute AI function with best intention for the player
	elseif sc_DIFF == -1 then
		gotoAI_1(-1)
		
	-- difficulty = 1, execute AI function with worst intention for the player
	elseif sc_DIFF == 1 then
		gotoAI_1(1)
	end
	
	
end

-- boolean used to turn real time AI simulation on/off
simMode = false

-- float specifying the speed of the AI real time display of the simulation
simSpeed = 0

-- timer used for simulation purposes
simTimer = 0

-- calculate fitness when simulating runs
function calcFitness()
				
		-- calculate the fitness value for this final board position.
		newF = fitness(newGrid, true)
				
		-- if this value is lower than a previous fitness value for the same piece,
		-- overwrite it
		if newF < V[iM] then
		
			-- update vector which check if the current piece has a chance to
			-- complete a line on the board
			if f6 then LC[iM] = true else LC[iM] = false end			
			
			V[iM] = newF
		end
		
		jM = jM + 1
		
		if jM == nCols or reachedEdge then
			jM = 0
			kM = kM + 1
		end
		
		if kM == 5 then
			kM = 1
			iM = iM + 1
		end
		
		if iM == 1 or iM == 3 or iM == 4 then
			if kM == 3 then
				kM = 1
				jM = 0
				iM = iM + 1
			end					
		end
		
		if iM == 6 and kM == 2 then
			kM = 1
			jM = 0
			iM = iM + 1			
		end
		
		if iM < 8 then
		
		else 
			simMode = false
			-- next we loop through all our fitness values (seven values, which are the minimum
			-- values for each piece after testing out all of its rotations and position) in order
			-- to find either a minimum value (best piece) or a maximum value (worst best piece)
			minV = V[1]
			minL = 1
			for i = 2, 7 do

				-- if we have benevolent AI, we want the piece with the minimum best fitness
				if sc_DIFF < 0 then

					if V[i] < minV then
						minV = V[i]
						minL = i
					end
	
				-- with the maleficent AI we want the piece with the maximum best fitness.
				-- remember, we don't calculate the worst fitnesses, instead calculating the
				-- best fitness then taking the worst of that. a piece is as bad as its worst
				-- best position
				else
					if V[i] > minV then
						minV = V[i]
						minL = i
					end
				end		
			end

			-- after identifying which piece we want (either best or worst), we create it!
			createBlock("", minL)
			curBlock.active = true

			-- update value on the stats board
			if LC[minL] then d6 = "Yes" else d6 = "No" end
			
			-- revert grid to its initial condition
			newGrid = copyGrid(newGrid, grid)
		
		end

end

-- function that copies the contents of one grid to another
function copyGrid(gridO, gridC)
	
	-- hypothetical grid serving as a testbed. starts out identical to our
	-- gameplay grid. is reset after each loop to test a new position
	gridO = {}
	
	-- copy the content of our gameplay grid to our AI grid
	for y=1, #gridC do
		gridO[y] = {}
		for x=1, #gridC[y] do
			gridO[y][x] = gridC[y][x]
		end
	end
	
	return gridO
end

-- function that simulate the A.I. on the second grid
function simulateRun()

		-- initialize the AI testbed grid with the values of our main grid at each run
		-- in the pieces/positions/rotations loop
		newGrid = copyGrid(newGrid, grid)

		-- create block according to (i)
		createBlock("", iM)
		
		-- rotate according to (k)
		for h = 2, kM do
			if canRotate(newGrid) then
				rotate(newGrid)
			end
		end
		
		-- move piece to the utmost right
		while canMove(1, 0, newGrid) do
			move(1, 0, newGrid)					
		end
		
		reachedEdge = false
		
		-- move piece to the left, one by one, to accomodate for all possible
		-- column choice to be dropped from, according to (j)
		if canMove(-jM, 0, newGrid) then
			move(-jM, 0, newGrid)			
		end
		
		if not canMove(-1, 0, newGrid) then
			reachedEdge = true
		end
		
		-- now that we have a piece, rotated in some way, placed in some column,
		-- drop it all the way down to end up with a final board position at the
		-- end of the turn, one of all possible permutation we are going to test.
		-- note that the code doesn't take into account moving left/right during
		-- landing of the piece.
		while canMove(0, 1, newGrid) do		
			move(0, 1, newGrid)					
		end

end

-- convert grid to string
function grid2str(G)
	
	str = ""
	for y = 1, #G do
		for x = 1, #G[y] do
			str = str..string.format('%i', G[y][x])
		end
	end
	
	return str
end

-- AI function. This function plays out all possible positions of all pieces in all
-- rotations, then it gives each position a fitness value and based on intention creates
-- the peace with either the best, or the worst fitness value
function gotoAI_1(H)

	-- reset the values of our AI testbed grid to that of the main grid
	newGrid = copyGrid(newGrid, grid)

	-- initialize two seven element arrays; one for storing the fitness values for each
	-- piece (V), and one to indicate whether this piece can clear a line on the board (LC)
	V = {}
	LC = {}
	V1 = {}
	
	-- initialize the values. fitness values are very high by default, because we are
	-- interested in minimizing the fitness for each piece's best board location.
	-- technically we are calculating the board's fitness value according to each piece's
	-- final location in it. the lowest number corresponds to the best piece in the current
	-- context
	for i = 1, 7 do
		V[i] = 999999999
		V1[i] = 999999999
		LC[i] = false
	end
	
	iM = 1
	kM = 1
	jM = -1
	reachedEdge = false
		
	-- depending on whether we are going to display the A.I. thinking process
	if not canSimulate then
	
		-- loop three times, to take into account all possible pieces (i = 1, 7), in all
		-- possible rotations (k = 1, 4) and lowered from all possible columns (j = 0, nCols - 1)
		for i = 1, 7 do
			for k = 1, 4 do
				for j = 0, nCols - 1, 1 do
					
					iM = i
					kM = k
					jM = j
					
					simulateRun()
					
					for i = 1, 4 do
	
						-- update second grid, make the matrix's element match the block's type
						newGrid[curBlock.co[i].y][curBlock.co[i].x] = curBlock.btype		
						
					end		
	
					calcFitness()
				end
		
			end
		end
			
		newGrid = {}
		
		-- copy the content of our gameplay grid to our AI grid
		for y=1, #grid do
			newGrid[y] = {}
			for x=1, #grid[y] do
				newGrid[y][x] = 0
			end
		end

	else
		simMode = true
	end
	
end

-- display stats, the same ones used to evaluate boards and their fitness values
function dispStats()
	
	-- maximum height by any block
	d1 = maxHeight(grid)
	
	-- average heights of all block
	d2 = meanHeight(grid)
	
	-- number of cavities and semi cavities on the board. a cavity is when four blocks
	-- surround an empty space. a semi cavity is when three blocks surround one
	d8, d3, d4 = cavities(grid)
	
	-- the number of block edges not in contact with any other block or border
	d5 = freeEdges(grid)
	
	-- based on the aforementioned metrics, a total fitness value may be deduced
	d7 = fitness(grid, false)
	
end

-- calculate the fitness of a board at a certain state, and by extension, the last piece
-- played on it
function fitness(grid, isAI)
	
	-- if we don't have an AI in place, we won't calculate whether a piece could
	-- complete a line or not. we can do it, but it's not as important
	if sc_DIFF == 0 then d6 = "NA" end

	-- if we're only here for the purpose of graphically displaying stats, just multiply
	-- each factor with its coefficient/weight and return the "fitness" value
	if not isAI then
		
		return c1 * d1
		 + c2 * d2
		  + c3 * d3
		   + c4 * d4
			+ c5 * d5
			 + c8 * d8
	end
	 
	-- if we're here to assess a board's fitness value after the drop of a piece on it,
	-- we calculate each of its metrics. these are the same criteria as the ones used in
	-- dispStats()

	-- maximum height by any block
	f1 = maxHeight(newGrid)
	
	-- average heights of all block
	f2 = meanHeight(newGrid)
	
	-- number of cavities and semi cavities on the board. a cavity is when four blocks
	-- surround an empty space. a semi cavity is when three blocks surround one. also
	-- obtains how many holes are closed only from the top and the bottom (f8)
	f8, f3, f4 = cavities(newGrid)
	
	-- the number of block edges not in contact with any other block or border
	f5 = freeEdges(newGrid)
	
	-- check if a current piece may complete and clear a full line
	f6 = doesLine(newGrid)
	
	-- based on these we calculate a final fitness value and return it
	
	if f6 then 
		w6 = 1
	else  
		w6 = 0
	end

	return c1 * f1
	+ c2 * f2
	+ c3 * f3
	+ c4 * f4
	+ c5 * f5
	+ c6 * w6
	+ c8 * f8
	
end

-- calculate the maximum height occupied by any block on the board
function maxHeight(xgrid)
	
	-- loop through all rows
	for i = 5, nRows + 4 do
		
		-- take the first row wherein its elements add up to a value more than 0
		local sum = 0
		for j = 1, nCols do
			sum = sum + xgrid[i][j]
		end
		
		-- this is our maximum height
		if sum > 0 then
			return nRows + 5 - i
		end
		
	end
	
	return 0
	
end

-- get the average height of all blocks on the board
function meanHeight(xgrid)

	-- loop through all rows, getting a sum of how many elements in each row are not 0
	-- loop through all elements, getting a total count of how many elements they are
	-- the mean is equal to the sum of all heights divided by how many blocks there are

	sum = 0
	tot = 0
	for i = 5, nRows + 4 do
		for j = 1, nCols do
			if xgrid[i][j] > 0 then
				sum = sum + xgrid[i][j]/xgrid[i][j] * (nRows + 5 - i)
				tot = tot + xgrid[i][j]/xgrid[i][j]

			end
		end		
	end
	
	if tot > 0 then
		return sum/tot
	else
		return 0	
	end
	
end

-- calculate how many holes are present on the board.
function cavities(xgrid)

	-- used to store holes closed from the top and bottom
	sumc = 0
	
	-- used to store holes surrounded on three sides
	sum = 0
	
	-- used to store holes surrounded on all four sides
	sumf = 0
	
	-- loop through all rows and columns
	for i = 5, nRows + 4 do
		for j = 1, nCols do
			
			-- if there is an empty space encountered
			if xgrid[i][j] == 0 then
			
				counter = 0				
			
				-- used to test if only closed from the top and bottom
				condr = false
				
				-- check if surrounded from above
				if i == 5 then
					counter = counter + 1
				elseif xgrid[i - 1][j] > 0 then
					counter = counter + 1
				end
				
				-- check if surrounded from below
				if i == nRows + 4 then
					counter = counter + 1
				elseif xgrid[i + 1][j] > 0 then
					counter = counter + 1
				end				
				
				-- only if closed from above and below is it eligible to be a cavity 
				if counter == 2 then
					condr = true
				end	
			
				-- check if surrounded from the right
				if j == nCols then
					counter = counter + 1
				elseif xgrid[i][j + 1] > 0 then
					counter = counter + 1
				end
			
				-- check if surrounded from the left
				if j == 1 then
					counter = counter + 1
				elseif xgrid[i][j - 1] > 0 then
					counter = counter + 1
				end
			
				-- depending on the number of sides it is surrounded by, either add it
				-- to the cavity counter, or the semi-cavity one
				if counter == 3 and condr then
					sum = sum + 1
				elseif counter == 4 and condr then
					sumf = sumf + 1
				elseif condr then
					sumc = sumc + 1
				end
				
			end
		end		
	end

	return sumc, sum, sumf
end

-- calculate the number of edges for all blocks that are not touching any other blocks
-- or borders
function freeEdges(xgrid)
	
	sum = 0
	
	-- loop through all elements
	for i = 5, nRows + 4 do
		for j = 1, nCols do
			
			-- if a block is present
			if xgrid[i][j] > 0 then
			
				-- check if there is anything to its top/down/left/right.
				-- similar to cavity check
				if i > 5 and xgrid[i - 1][j] == 0 then sum = sum + 1 end
				if i < nRows + 4 and xgrid[i + 1][j] == 0 then sum = sum + 1 end
				if j > 1 and xgrid[i][j - 1] == 0 then sum = sum + 1 end
				if j < nCols and xgrid[i][j + 1] == 0 then sum = sum + 1 end
			end
		end
	end
	
	return sum
end

-- check if current piece on board resulted in a full row of blocks
function doesLine(xgrid)
	
	-- loop through all rows.
	for i = 5, nRows + 4 do
	
		-- check if x > 0 holds true for all columns in a row
		local sum = true
		for j = 1, nCols do
			sum = sum and xgrid[i][j] > 0
		end
		
		-- if yes, then we have a full row of blocks. similar to checkLines()
		if sum then
			return true
		end
	end

	return false

end

-- function that pauses game on loss/return of focus
function love.focus(f)
	gameIsPaused = true
end

-- save score on exit. not necessary since score is saved at the end of each turn
function love.quit()
	-- love.filesystem.write(fname, sc_SCORE)
end

return maingame